"""Tests for the queue model."""
from talus.models.queue import Queue


def test_queue():
    """
    :given: A queue name.
    :when: A queue is created.
    :then: The queue name is accessible with expected defaults.
    """
    # given
    queue_name = "foo"
    # when
    queue = Queue(name=queue_name)
    # then
    assert queue.name == queue_name
    assert queue.passive is False
    assert queue.durable is True
    assert queue.auto_delete is False
    assert queue.exclusive is False
    assert queue.arguments is None



"""Tests for the message models."""
import json
from typing import Any

import pika
import pydantic_core
import pytest

from talus.models.message import ConsumeMessageBase
from talus.models.message import MessageBodyBase
from talus.models.message import PublishMessageBase


def test_consume_message_model_valid(
    method, properties, body, routing_key, headers, delivery_tag, message_data
):
    """
    :given: A ConsumeMessage class
    :when: A new instance is created with default args
    :then: The instance body has a conversationId and other properties are set
    """
    # given/when
    message = ConsumeMessageBase(method=method, properties=properties, body=body)
    # then
    assert message.body.conversationId is not None
    assert message.method == method
    assert message.properties == properties
    assert message.routing_key == routing_key
    assert message.headers == headers
    assert message.delivery_tag == delivery_tag
    assert (
        message.body.model_dump() == {"conversationId": message.body.conversationId} | message_data
    )
    assert repr(message)


class ExampleMessageBody(MessageBodyBase):
    required_str: str
    required_int: int


class ConsumableMessage(ConsumeMessageBase):
    message_body_cls = ExampleMessageBody


@pytest.mark.parametrize(
    "invalid_body",
    [
        pytest.param(
            json.dumps({"required_str": 1, "required_int": "zz"}).encode("utf-8"),
            id="schema_invalid",
        ),
        pytest.param(b"123 abc", id="json_invalid"),
    ],
)
def test_consume_message_model_invalid(method, properties, invalid_body: Any):
    """
    :given: A ConsumeMessage class
    :when: A new instance is created and the body is invalid for the message schema
    :then: A Validation Error is raised
    """
    with pytest.raises(pydantic_core.ValidationError):
        ConsumableMessage(method=method, properties=properties, body=invalid_body)


def test_publish_message_model(routing_key, headers, body, message_data, publish_message_cls):
    """
    :given: A PublishMessage class
    :when: A new instance is created
    :then: The instance body has a conversationId and other properties are set
    """
    # given/when
    message = publish_message_cls(body=body)
    # then
    assert message.body.conversationId is not None
    assert message.routing_key == routing_key
    assert message.headers == headers
    assert isinstance(message.properties, pika.spec.BasicProperties)
    assert message.properties.headers == headers
    assert (
        message.body.model_dump() == {"conversationId": message.body.conversationId} | message_data
    )
    assert repr(message)


class PublishMessage(PublishMessageBase):
    message_body_cls = ExampleMessageBody
    default_routing_key = "test.m"


@pytest.mark.parametrize(
    "invalid_body",
    [
        pytest.param(
            json.dumps({"required_str": 1, "required_int": "zz"}).encode("utf-8"),
            id="schema_invalid",
        ),
        pytest.param(b"123 abc", id="json_invalid"),
    ],
)
def test_publish_message_model_invalid(routing_key, invalid_body):
    """
    :given: A PublishMessage class
    :when: A new instance is created and the body is invalid for the message schema
    :then: A Validation Error is raised
    """
    with pytest.raises(pydantic_core.ValidationError):
        PublishMessage(body=invalid_body)



"""Test the Retryer model."""
from tenacity import Retrying

from talus.models.retryer import ConnectionRetryerFactory
from talus.models.retryer import DEFAULT_CONNECTION_EXCEPTIONS
from talus.models.retryer import RetryerFactory


def test_retryer_defaults():
    """
    :given: A retryer factory
    :when: The factory is called
    :then: A Retryer object is returned
    """
    # given
    factory = RetryerFactory()
    # when
    retryer = factory()
    # then
    assert isinstance(retryer, Retrying)


def test_retryer_with_attempts():
    """
    :given: A retryer factory with attempts > 1
    :when: The factory is called
    :then: A Retryer object is returned with a stop_after_attempt object
    """
    # given
    factory = RetryerFactory(attempts=5)
    # when
    retryer = factory()
    # then
    assert isinstance(retryer, Retrying)
    assert retryer.stop is not None


def test_connection_retryer_factory():
    """
    :given: A connection retryer factory
    :when: The factory is called
    :then: A Retrying object is returned
    """
    # given
    factory = ConnectionRetryerFactory()
    # when
    retryer = factory()
    # then
    assert isinstance(retryer, Retrying)
    assert factory.exceptions == DEFAULT_CONNECTION_EXCEPTIONS



"""Tests for the durable producer module."""
from talus.producer import DurableProducer


def test_publish(producer, publish_message_cls, message_data, consumer, consume_message_cls):
    """
    :given: A producer and message data
    :when: The producer publishes a message
    :then: The message is published
    """
    # when
    publish_message = publish_message_cls(message_data)
    producer.publish(publish_message)
    # then
    for method, properties, body in consumer.consume_generator(auto_ack=True):
        message = consume_message_cls(method=method, properties=properties, body=body)
        for field_name in message.body.model_fields:
            assert getattr(message.body, field_name) == getattr(publish_message.body, field_name)
        break


def test_repr(producer):
    """
    :given: A producer
    :when: The producer is represented as a string
    :then: The string representation is as expected
    """
    assert DurableProducer.__name__ in repr(producer)



"""
Bindings contain the many-to-many relationship between a message routing key and a queue on an exchange.
"""
from typing import Type

from pydantic import AfterValidator
from pydantic import BaseModel
from pydantic import ConfigDict
from typing_extensions import Annotated

from talus.models.message import PublishMessageBase
from talus.models.queue import Queue


def validate_routing_key(message: Type[PublishMessageBase]):
    """Custom pydantic validator to ensure that the routing key is not empty."""
    if not message.default_routing_key:
        raise ValueError("default_routing_key must not be empty")
    return message


class Binding(BaseModel):
    """
    Binding configuration.

    >>> from talus.models.binding import Binding
    >>> from talus.models.message import PublishMessageBase
    >>> from talus.models.queue import Queue
    >>> queue = Queue(name="my.queue")
    >>> class PublishMessage(PublishMessageBase):
    >>>     default_routing_key = "my.routing.key"
    >>> binding = Binding(message=PublishMessage, queue=queue)
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    message: Annotated[Type[PublishMessageBase], AfterValidator(validate_routing_key)]
    queue: Queue

    @property
    def routing_key(self) -> str:
        return self.message.default_routing_key

    @property
    def queue_name(self) -> str:
        return self.queue.name



"""Queue model."""
from pydantic import BaseModel


class Queue(BaseModel):
    """
    Queue model.
    >>> from talus.models.queue import Queue
    >>> queue = Queue(name="my.queue")
    """

    name: str
    passive: bool = False
    durable: bool = True
    auto_delete: bool = False
    exclusive: bool = False
    arguments: dict | None = None

    def __str__(self):
        return self.name



"""Tests for the durable consumer module."""
from talus.consumer import DurableConsumer


def test_consumer_listen(consumer, single_message_on_queue, message_data, consume_message_cls):
    """
    :given: A consumer and a message on the queue
    :when: The consumer listens for a message
    :then: The message is consumed
    """

    def verify_message(ch, method, properties, body):  # then
        message = consume_message_cls(method=method, properties=properties, body=body)
        for field_name in message.body.model_fields:
            assert getattr(message.body, field_name) == getattr(
                single_message_on_queue.body, field_name
            )
        ch.basic_ack(method.delivery_tag)
        consumer.disconnect()

    # when
    consumer.listen(verify_message)


def test_consume_generator(consumer, single_message_on_queue, consume_message_cls):
    """
    :given: A consumer and a message on the queue
    :when: The consumer consumes a message via consume_generator
    :then: The expected message is consumed and the generator cancelled
    """
    # when
    for method, properties, body in consumer.consume_generator():
        # then
        message = consume_message_cls(method=method, properties=properties, body=body)
        for field_name in message.body.model_fields:
            assert getattr(message.body, field_name) == getattr(
                single_message_on_queue.body, field_name
            )
        consumer.cancel_consume_generator()


def test_acknowledge_message(consumer, single_message_on_queue, consume_message_cls):
    """
    :given: A consumer and a message on the queue
    :when: The consumer acknowledges a message
    :then: The message is consumed
    """
    # when
    for method, properties, body in consumer.consume_generator():
        message = consume_message_cls(method=method, properties=properties, body=body)
        for field_name in message.body.model_fields:
            assert getattr(message.body, field_name) == getattr(
                single_message_on_queue.body, field_name
            )
        consumer.acknowledge_message(method.delivery_tag)
        break
    # then
    for method, properties, body in consumer.consume_generator():
        assert body is None
        break


def test_reject_message(consumer, single_message_on_queue, consume_message_cls):
    """
    :given: A consumer and a message on the queue
    :when: The consumer rejects a message
    :then: The message is not consumed
    """
    # when
    for method, properties, body in consumer.consume_generator():
        message = consume_message_cls(method=method, properties=properties, body=body)
        for field_name in message.body.model_fields:
            assert getattr(message.body, field_name) == getattr(
                single_message_on_queue.body, field_name
            )
        consumer.reject_message(message.delivery_tag)
        break
    # then
    for method, properties, body in consumer.consume_generator():
        assert body is None
        break


# requeue
def test_requeue_message(consumer, single_message_on_queue, consume_message_cls):
    """
    :given: A consumer and a message on the queue
    :when: The consumer requeues a message
    :then: The message is requeued
    """
    for _ in range(2):
        # when (loop 0) / then(loop 1)
        for method, properties, body in consumer.consume_generator():
            message = consume_message_cls(method=method, properties=properties, body=body)
            for field_name in message.body.model_fields:
                assert getattr(message.body, field_name) == getattr(
                    single_message_on_queue.body, field_name
                )
            consumer.requeue_message(method.delivery_tag)
            break


def test_repr(consumer):
    """
    :given: A consumer
    :when: The consumer is represented as a string
    :then: The string representation is returned
    """
    assert DurableConsumer.__name__ in repr(consumer)



"""
Module encapsulating message structure to facilitate use with the Consumer and Producer wrappers
"""
import uuid
from typing import Type

import pika.spec
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field


class MessageBodyBase(BaseModel):
    """
    Base class for message schemas which can be used to validate message bodies.
    This is a pydantic.BaseModel so each message body schema can be validated.  Extend this class
    to add fields to a new message body schema.  For Example:
    >>> class ObjectBucketBody(MessageBodyBase):
    >>>     objectName: str
    >>>     bucket: str
    """

    model_config = ConfigDict(extra="allow")

    conversationId: str = Field(default_factory=lambda: uuid.uuid4().hex)


class _MessageBase:
    """
    Base class for messages establishing a common interface for use by DurableConnections.
    Not intended for direct use.  Use ConsumeMessageBase or PublishMessageBase instead.
    """

    message_body_cls: Type[MessageBodyBase] = MessageBodyBase

    def __init__(self, routing_key: str, body: bytes | dict | str | MessageBodyBase):
        self.routing_key = routing_key
        if isinstance(body, dict | BaseModel):
            self.body: MessageBodyBase = self.message_body_cls.model_validate(body)
        else:  # str | bytes
            self.body: MessageBodyBase = self.message_body_cls.model_validate_json(body)


class ConsumeMessageBase(_MessageBase):
    """
    Base class for messages consumed from a DurableConsumer MessageProcessor.  Each instance
    correlates with a single consumed message.  The body of the message is validated against the
    message_body_cls class attribute.
    >>> class ObjectBucketBody(MessageBodyBase):
    >>>    objectName: str
    >>>    bucket: str
    >>>
    >>> class InboundMessage(ConsumeMessageBase):
    >>>     message_body_cls: Type[ObjectBucketBody] = ObjectBucketBody  # Validate the body with ObjectBucketBody
    >>>
    >>> queue = Queue("my_queue")
    >>> with DurableConsumer(consume_queue=queue) as consumer:
    >>>     consumer.consume(lambda ch, method, properties, body: InboundMessage(method, properties, body))
    """

    def __init__(
        self,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes | dict | str | MessageBodyBase,
    ):
        super().__init__(method.routing_key, body)
        self.method = method
        self.properties = properties

    @property
    def delivery_tag(self):
        return self.method.delivery_tag

    @property
    def headers(self) -> dict | None:
        return self.properties.headers

    def __repr__(self):
        return f"{self.__class__.__name__}(method={self.method!r}, properties={self.properties!r}, body={self.body!r})"


class PublishMessageBase(_MessageBase):
    """
    Base class for messages published to with a DurableProducer.
    Each instance correlates with a single message to publish.  The body of the message is validated
    against the class defined in message_body_cls class attribute.  Additional class attributes for
    routing_key and headers are provided for customizing the message at the class level e.g. all
    instances share a routing_key.

    >>> class OutboundMessage(PublishMessageBase):
    >>>     message_body_cls: Type[ObjectBucketBody] = ObjectBucketBody
    >>>     default_routing_key: str = "outbound.m"
    >>>     headers: dict[str, str] = Field(default={"header1": "value1"})
    >>>
    >>>
    >>> queue = Queue(name="test_queue")
    >>> exchange = Exchange(name="test_exchange")
    >>> queue_bindings = Binding(queue=queue, message=OutboundMessage)
    >>> with DurableProducer(queue_bindings=queue_bindings, publish_exchange=exchange) as producer:
    >>>     body = {"objectName": "object", "bucket": "bucket"}
    >>>     producer.publish(OutboundMessage(body))
    """

    default_routing_key: str = "default.m"
    headers: dict[str, str] | None = None

    def __init__(
        self,
        body: bytes | dict | str | MessageBodyBase,
    ):
        super().__init__(self.default_routing_key, body)

    @property
    def properties(self) -> pika.spec.BasicProperties:
        return pika.BasicProperties(
            content_type="text/plain",
            priority=0,
            delivery_mode=pika.DeliveryMode.Persistent,
            content_encoding="UTF-8",
            headers=self.headers,
        )

    def __repr__(self):
        return f"{self.__class__.__name__}(body={self.body!r}) # {self.routing_key = }"



"""
Consumer implementation of the base connection wrapper
"""
import logging
from typing import Callable
from typing import Generator

import pika.adapters.blocking_connection
from tenacity import Retrying

from talus.base import DurableConnection
from talus.models.connection_parameters import ConsumerConnectionParameterFactory
from talus.models.queue import Queue
from talus.models.retryer import ConnectionRetryerFactory

logger = logging.getLogger(__name__)


LISTEN_CALLBACK_TYPE = Callable[
    [
        pika.adapters.blocking_connection.BlockingChannel,
        pika.spec.Basic.Deliver,
        pika.spec.BasicProperties,
        bytes,
    ],
    None,
]


class DurableConsumer(DurableConnection):
    """
    RabbitMQ connector for consuming from a single queue in RabbitMQ.
    >>> from talus.models.queue import Queue
    >>> from talus.models.connection_parameters import ConsumerConnectionParameterFactory
    >>> from talus.models.retryer import ConnectionRetryerFactory
    >>> from talus.consumer import DurableConsumer
    >>> from talus.models.processor import MessageProcessorBase
    >>> from talus.models.message import ConsumeMessageBase
    >>> consume_queue = Queue(name="test_queue")
    >>> class MessageProcessor(MessageProcessorBase):
    >>>     def process_message(message: ConsumeMessageBase):
    >>>         print(message)
    >>> with DurableConsumer(consume_queue=consume_queue, prefetch_count=1, connection_parameters=ConsumerConnectionParameterFactory(), connection_retryer=ConnectionRetryerFactory()) as consumer:
    >>>     with MessageProcessor() as message_processor:
    >>>         consumer.listen(message_processor)
    """

    def __init__(
        self,
        consume_queue: Queue,
        prefetch_count: int = 1,
        connection_parameters: pika.ConnectionParameters
        | Callable[[], pika.ConnectionParameters] = ConsumerConnectionParameterFactory(),
        connection_retryer: Retrying | Callable[[], Retrying] = ConnectionRetryerFactory(),
    ):
        """
        Constructor for the consumer connector

        :param consume_queue: Queue to consume messages from

        :param prefetch_count: Number of un-Acked message delivered at a time

        :param connection_parameters: A pika.ConnectionParameters object or one resulting
            from a callable that returns a ConnectionParameters object.

        :param connection_retryer: A tenacity.Retrying object or one resulting from a callable that returns
            a Retrying object.
        """
        super().__init__(
            connection_parameters=connection_parameters,
            connection_retryer=connection_retryer,
        )
        self.consume_queue = consume_queue
        self.prefetch_count = prefetch_count

    def _connect(self) -> None:
        """
        Configures and initiates consumer connection to the RabbitMQ server which includes
        setting up the queue to consume from and the prefetch count.
        """
        super()._connect()
        self.channel.basic_qos(prefetch_count=self.prefetch_count)
        self.create_queue(queue=self.consume_queue)

    def _listen(self, callback: LISTEN_CALLBACK_TYPE) -> None:
        """
        Listens for messages on the channel configured on the consumer instance

        :param callback: Function to execute when a message is received. with the signature
        (ch, method, properties, body).
        ch: Copy of the channel used to acknowledge receipt (pika.Channel)
        method: Management keys for the delivered message e.g. delivery mode (pika.spec.Basic.Deliver)
        properties: Message properties (pika.spec.BasicProperties)
        body: Message body for a transfer message (bytes)
        """
        self.connect()
        logger.info(f"Starting Listener on Queue: consumer_queue={self.consume_queue}")
        self.channel.basic_consume(queue=self.consume_queue.name, on_message_callback=callback)
        self.channel.start_consuming()

    def listen(self, message_processor: LISTEN_CALLBACK_TYPE) -> None:
        """
        Retries calls to _listen and executes the message_processor callback when a message is received.
        This method is blocking and will not return until the connection is closed.
        :param message_processor: Callable to execute when a message is received. with the signature
        (channel, method, properties, body).
        channel: Copy of the channel used to acknowledge receipt (pika.Channel)
        method: Management keys for the delivered message e.g. delivery mode (pika.spec.Basic.Deliver)
        properties: Message properties (pika.spec.BasicProperties)
        body: Message body for a transfer message (bytes)
        and returns None.
        """
        self.connection_retryer(self._listen, callback=message_processor)

    def consume_generator(self, auto_ack=False, inactivity_timeout: float = 0.1) -> Generator:
        """
        Creates a generator for messages that are on the instance consumer_queue.
        Retry logic is not applied to prevent the resetting of the generator cursor

        :param auto_ack: Automatically acknowledge messages
        :param inactivity_timeout: Number of seconds to wait for a message before returning None

        :return: Generator of (method, properties, body)
        """
        self.connect()
        logger.info(f"Creating consumer generator on Queue: consumer_queue={self.consume_queue}")
        return self.channel.consume(
            queue=self.consume_queue.name, auto_ack=auto_ack, inactivity_timeout=inactivity_timeout
        )

    def cancel_consume_generator(self) -> None:
        """
        Resets the active consume generator
        :return: None
        """
        logger.info(f"Cancelling consumer generator on Queue: consumer_queue={self.consume_queue}")
        self.channel.cancel()

    def acknowledge_message(self, delivery_tag, multiple=False) -> None:
        """
        Record a message as acknowledged.
        Retry logic is not applied since creating a new channel would be unable
        to acknowledge the message received on the now dead channel

        :param delivery_tag: method.delivery_tag

        :param multiple: Acknowledge multiple messages by setting to True and acknowledging the last message

        :return: None
        """
        self.channel.basic_ack(delivery_tag, multiple)

    def reject_message(self, delivery_tag) -> None:
        """
        Record a message as rejected.  Will go to dead letter exchange if configured on the server.
        Retry logic is not applied since creating a new channel would be unable
        to use the delivery tag received on the now dead channel

        :param delivery_tag: method.delivery_tag

        :return: None
        """
        self.channel.basic_reject(delivery_tag=delivery_tag, requeue=False)

    def requeue_message(self, delivery_tag) -> None:
        """
        Return message back to the queue.
        Retry logic is not applied since creating a new channel would be unable
        to use the delivery tag received on the now dead channel

        :param delivery_tag: method.delivery_tag

        :return: None
        """
        self.channel.basic_nack(delivery_tag=delivery_tag, requeue=True)

    def __repr__(self):
        """Representation of the DurableConsumer instance."""
        return f"{self.__class__.__name__}(consume_queue={self.consume_queue !r}, prefetch_count={self.prefetch_count}, connection_parameters={self.connection_parameters!r}, connection_retryer={self.connection_retryer!r})"



"""Tests for the message processor models."""
from typing import Type
from uuid import uuid4

import pika
import pytest
from pydantic import Field

from talus.models.connection_parameters import ProducerConnectionParameterFactory
from talus.models.message import ConsumeMessageBase
from talus.models.processor import DLQMessageProcessor
from talus.models.processor import MessageProcessorBase
from talus.models.processor import MessageProcessorRoutingKeySelector
from talus.models.retryer import ConnectionRetryerFactory
from talus.producer import DurableProducer


@pytest.fixture()
def successful_message_processor(consume_message_cls) -> MessageProcessorBase:
    """
    A message processor that does not raise an exception.
    """

    class SuccessfulMessageProcessor(MessageProcessorBase):
        message_cls: Type[ConsumeMessageBase] = consume_message_cls

        def process_message(self, message: ConsumeMessageBase):
            self._message_was_processed = True

    yield SuccessfulMessageProcessor()


@pytest.fixture()
def consumed_message(
    publish_message_cls,
    producer,
    consumer,
    body,
) -> tuple[pika.spec.Basic.Deliver, pika.spec.BasicProperties, bytes]:
    outbound_message = publish_message_cls(body=body)
    producer.publish(message=outbound_message)
    for m, p, b in consumer.consume_generator(auto_ack=True):
        yield m, p, b  # only the first one
        break


def test_message_processor_success(consumer, consumed_message, successful_message_processor):
    """
    :given: A message processor that does not raise an exception.
    :when: The message processor is called.
    :then: The message processor should process and acknowledge the message.
    """
    # given
    method, properties, body = consumed_message
    # when
    successful_message_processor(
        channel=consumer.channel, method=method, properties=properties, body=body
    )
    # then
    assert successful_message_processor._message_was_processed
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue


@pytest.fixture()
def erroring_message_processor(consume_message_cls) -> MessageProcessorBase:
    """
    A message processor that raises a non-dead-letter exception.
    """

    class ErroringMessageProcessor(MessageProcessorBase):
        message_cls: Type[ConsumeMessageBase] = consume_message_cls
        dead_letter_exceptions: Type[Exception] = ValueError

        def process_message(self, message: ConsumeMessageBase):
            raise RuntimeError("A bad message processor raised an exception that is not DLQ'd.")

    yield ErroringMessageProcessor()


def test_message_processor_failure(erroring_message_processor, consumed_message, consumer):
    """
    :given: A message processor that raises an exception.
    :when: The message processor is called.
    :then: Expected exception is raised.
    """
    # given
    method, properties, body = consumed_message
    # when/then
    with pytest.raises(RuntimeError):
        erroring_message_processor(
            channel=consumer.channel, method=method, properties=properties, body=body
        )


@pytest.fixture()
def dlq_message_processor(consume_message_cls) -> DLQMessageProcessor:
    """
    A message processor that raises a non-dead-letter exception.
    """

    class AutoDLQMessageProcessor(DLQMessageProcessor):
        message_cls: Type[ConsumeMessageBase] = consume_message_cls
        dead_letter_exceptions: Type[Exception] = RuntimeError

        def process_message(self, message: ConsumeMessageBase):
            self._raised_run_time_error = True
            raise RuntimeError("A bad message processor raised an exception that is DLQ'd.")

    yield AutoDLQMessageProcessor()


def test_dlq_message_processor(dlq_message_processor, consumed_message, consumer):
    """
    :given: A message processor that raises an exception identified as for DLQ.
    :when: The message processor is called.
    :then: The message processor should DLQ the message.
    """
    # given
    method, properties, body = consumed_message
    # when
    dlq_message_processor(channel=consumer.channel, method=method, properties=properties, body=body)
    # then
    assert dlq_message_processor._raised_run_time_error
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue


@pytest.fixture()
def message_processor_routing_key_selector(
    routing_key, successful_message_processor
) -> MessageProcessorRoutingKeySelector:
    """
    A message processor selector that selects a message processor based on the routing key.
    """
    return MessageProcessorRoutingKeySelector(
        message_processors={
            routing_key: successful_message_processor,
        }
    )


def test_message_processor_routing_key_selector_found(
    message_processor_routing_key_selector, consumed_message, consumer, successful_message_processor
):
    """
    :given: A message processor selector that selects a message processor based on the routing key.
    :when: The message processor selector is called with a message of that routing key.
    :then: The expected message processor should be run and returned.
    """
    # given
    method, properties, body = consumed_message
    # when
    message_processor = message_processor_routing_key_selector(
        channel=consumer.channel, method=method, properties=properties, body=body
    )
    # then
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue
    assert isinstance(message_processor, type(successful_message_processor))


def test_message_processor_routing_key_selector_not_found(
    message_processor_routing_key_selector, consumer, consumed_message
):
    """
    :given: A message processor selector that selects a message processor based on the routing key.
    :when: The message processor selector is called with a message an unknown routing key.
    :then: The DLQ message processor should be returned.
    """
    # given
    method, properties, body = consumed_message
    unknown_routing_key = uuid4().hex
    method.routing_key = unknown_routing_key
    # when/then
    message_processor = message_processor_routing_key_selector(
        channel=consumer.channel, method=method, properties=properties, body=body
    )
    # then
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue
    assert isinstance(message_processor, DLQMessageProcessor)


@pytest.fixture()
def message_processor_cls_with_producer(producer, queue_bindings, direct_exchange):
    """
    A message processor that has a producer.
    """
    durable_producer = producer

    class MessageProcessorWithProducer(MessageProcessorBase):
        message_cls: Type[ConsumeMessageBase] = ConsumeMessageBase
        producer: DurableProducer = Field(default_factory=lambda: durable_producer)

        def process_message(self, message: ConsumeMessageBase):
            pass

    return MessageProcessorWithProducer
